#ifndef SCENE_H
#define SCENE_H

//-----------------------------------------------------------------------------
#include "Base.h"

#include <math.h>


//-----------------------------------------------------------------------------
class CObject;
class CLight;
class CRay;

//-----------------------------------------------------------------------------
/// scene class
/*
* 
*/
class CScene
{
//-----------------------------------------------------------------------------
private:

  /** maximal raytracing depth */
  int maxDepth;
  
  /** near clipping plane */
  double nearC;
  /** far clipping plane */
  double farC;

  /** background color */
  CColor background;

  /** the objects in this scene */
  std::vector < CObject* > objects;
  /** the lights in this scene */
  std::vector < CLight* > lights;

//-----------------------------------------------------------------------------
public:
 
  /** Ctor. */
  CScene( const CColor& backround, const double nearC, const double farC )
    : maxDepth( 0 ), nearC( nearC ), farC( farC ), background( backround )
  {};

  /** Dtor. */
  ~CScene();

  /** Return the background color. */
  const CColor& backgroundColor() const {
    return background;
  };

  /** Return the number of objects. */
  const unsigned int numObjects() const {
    return objects.size();
  };

  /** Add an object to the scene. */
  void addObject( CObject* object ) {
    objects.push_back( object );
  };

  /** Return the number of lights. */
  const int numLights() const {
    return (int)lights.size();
  };

  /** Add the given light to this scene. */
  void addLight( CLight* light ) {
    lights.push_back( light );
  };

  /** Get the light at the given index. */
  CLight& light( const int idx ) {
    return *lights[ idx ];
  };

  /** Get the light at the given index. */
  const CLight& light( const int idx ) const {
    return *lights[ idx ];
  };

  /** Return the raytracing depth. */
  const int maxRenderDepth() const
  {
    return maxDepth;
  };

  /** Increase the raytracing depth. */
  void increaseRenderDepth() {
    maxDepth++;
  };

  /** Decrease the raytracing depth. */
  void decreaseRenderDepth()
  {
	  if ( maxDepth < 1 )
		  maxDepth = 0;
	  else {
		  maxDepth --;
	  }
  };

  // Return the n-th object, n = 0, .., num. objects - 1
  CObject * getObject( unsigned int n ) const
  {
	  if ( n >= numObjects() )
		  return NULL;
	  return objects[n];
  };
	
  /** 
  * Set the near clipping plane to the given value.
  * @param newNearC - the new near clipping plane
  */
  void setNearClippingPlane( const double newNearC ) {
    nearC = newNearC;
  };

  /** 
  * Intersect the objects in the scene with the ray 
  * and returns the color at the intersection point.
  * @param ray - the ray
  * @param iter - the actual iteration step
  */
  const CColor intersection( const CRay& ray, const int iter = 0 ) const;

//-----------------------------------------------------------------------------
private:

  /**
  * Returns true if the ray intersects an object in this scene.
  * @param ray - the ray
  */
  const bool intersectionTest( const CRay& ray ) const;

  /**
  * Returns the color at the given intersection point.
  * @param object - the intersected object
  * @param ray - the ray
  * @param iP - the intersection point
  * @param iter - the actual iteration step
  */
  const CColor shading( const CObject* obj, const CRay& ray,
    const Point3d& iP, const int iter ) const;

};

#endif
